/*
 * Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the University of California, Berkeley nor
 *   the names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior 
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;

/**
 * Interface for objects that can choose a value for a basic random 
 * variable, given the applicable CPD and CPD arguments.
 */
public abstract class ValueChooser {
    /**
     * Returns a value for the given basic random variable.  May make use 
     * of the given CPD and CPD arguments.  May add an identifier to the 
     * given partial world, but does not set the value of any basic variable
     * -- the client must do that itself.
     */
    public abstract Object chooseValue(PartialWorld w, 
				       BasicVar var, 
				       CondProbDistrib cpd, 
				       List cpdArgs);

    /**
     * Returns a value for the given generating function on the given 
     * object.
     */
    public abstract Object chooseGeneratingObj(ObjectIdentifier generated, 
					       GeneratingFunction genFunc);

    /**
     * Adds a new identifier to the given world and sets the values of its 
     * generating functions so that it matches the given identifier hook.
     */
    public void addIdentifierForHook(PartialWorld w, IdentifierHook hook) {
	((MutablePartialWorld) w).addIdentifierForPOPApp
	    (hook.getPOP(), hook.getGeneratingObjects());
    }

    public static final ValueChooser NO_INSTANTIATION = 
	new ValueChooser() {
	    public Object chooseValue(PartialWorld w, BasicVar var, 
				      CondProbDistrib cpd, List cpdArgs) {
		throw new IllegalStateException
		    ("Variable " + var + " needs to be instantiated.");
	    }

	    public Object chooseGeneratingObj(ObjectIdentifier generated, 
					      GeneratingFunction genFunc) {
		throw new IllegalStateException
		    ("Generating function " + genFunc 
		     + " needs to be instantiated on " + generated);
	    }

	    public void addIdentifierForHook(PartialWorld w, 
					     IdentifierHook hook) {
		throw new IllegalStateException
		    ("Need to evaluate function on non-guaranteed object "
		     + "that has no identifier (" + hook + ")");
	    }
	};

}
